// taken from PixiJS
// https://pixijs.download/v8.0.0-rc.2/docs/rendering_renderers_gpu_texture_utils_GpuMipmapGenerator.ts.html

import code from './mip-shader.wgsl'

const mip = (n: number) => Math.max(1, n >>> 1)

export class MipGenerator {
  _device: GPUDevice
  _sampler: GPUSampler
  _pipelines: Record<string, GPURenderPipeline>

  _shader?: GPUShaderModule
  _bindGroupLayout?: GPUBindGroupLayout
  _pipelineLayout?: GPUPipelineLayout

  constructor(device: GPUDevice) {
    this._device = device
    this._sampler = device.createSampler({ minFilter: 'linear' })
    this._pipelines = {}
  }

  _getShader() {
    if (!this._shader) {
      this._shader = this._device.createShaderModule({
        code
      })
    }

    return this._shader
  }

  _getBindGroupLayout() {
    if (!this._bindGroupLayout) {
      this._bindGroupLayout = this._device.createBindGroupLayout({
        entries: [{
          binding: 0,
          visibility: GPUShaderStage.FRAGMENT,
          sampler: {}
        }, {
          binding: 1,
          visibility: GPUShaderStage.FRAGMENT,
          texture: {}
        }]
      })
    }

    return this._bindGroupLayout
  }

  _getPipelineLayout() {
    if (!this._pipelineLayout) {
      this._pipelineLayout = this._device.createPipelineLayout({
        label: 'Mipmap Generator',
        bindGroupLayouts: [this._getBindGroupLayout()],
      })
    }

    return this._pipelineLayout
  }

  getPipeline(format: GPUTextureFormat) {
    let pipeline = this._pipelines[format]

    if (!pipeline) {
      const shader = this._getShader()

      pipeline = this._device.createRenderPipeline({
        layout: this._getPipelineLayout(),
        vertex: {
          module: shader,
          entryPoint: 'vs_main'
        },
        fragment: {
          module: shader,
          entryPoint: 'fs_main',
          targets: [{ format }]
        }
      })

      this._pipelines[format] = pipeline
    }

    return pipeline
  }

  generateMipmap(texture: GPUTexture, commandEncoder?: GPUCommandEncoder): GPUTexture {
    const pipeline = this.getPipeline(texture.format)

    const { dimension, format, mipLevelCount, width, height, depthOrArrayLayers } = texture

    const encoder = commandEncoder || this._device.createCommandEncoder()

    if (texture.dimension !== '2d') {
      throw new Error('Generating mipmaps for anything except 2d is unsupported.')
    }

    let tempTexture = texture


    const renderToSource = texture.usage & GPUTextureUsage.RENDER_ATTACHMENT

    if (!renderToSource) {
      tempTexture = this._device.createTexture({
        size: {
          width: mip(width),
          height: mip(height),
          depthOrArrayLayers
        },
        format,
        usage: GPUTextureUsage.TEXTURE_BINDING | GPUTextureUsage.COPY_SRC | GPUTextureUsage.RENDER_ATTACHMENT,
        mipLevelCount: mipLevelCount - 1,
        dimension,
        sampleCount: 1
      })
    }

    for (let layer = 0; layer < depthOrArrayLayers; ++layer) {
      let srcView = texture.createView({
        baseMipLevel: 0,
        mipLevelCount: 1,
        dimension: '2d',
        baseArrayLayer: layer,
        arrayLayerCount: 1
      })

      for (let i = 1; i < mipLevelCount; ++i) {
        const dstView = tempTexture.createView({
          baseMipLevel: i - (renderToSource ? 0 : 1),
          mipLevelCount: 1,
          dimension: '2d',
          baseArrayLayer: layer,
          arrayLayerCount: 1
        })

        const passEncoder = encoder.beginRenderPass({
          colorAttachments: [{
            view: dstView,
            loadOp: 'clear',
            storeOp: 'store'
          }]
        })

        const bindGroup = this._device.createBindGroup({
          layout: this._bindGroupLayout,
          entries: [{
            binding: 0,
            resource: this._sampler
          }, {
            binding: 1,
            resource: srcView
          }]
        })

        passEncoder.setPipeline(pipeline)
        passEncoder.setBindGroup(0, bindGroup)
        passEncoder.draw(3, 1, 0, 0)
        passEncoder.end()

        srcView = dstView
      }
    }

    if (!renderToSource) {
      const mipSize = {
        width: mip(width),
        height: mip(height),
        depthOrArrayLayers
      }

      for (let i = 1; i < texture.mipLevelCount; ++i) {
        encoder.copyTextureToTexture({
          texture: tempTexture,
          mipLevel: i - 1
        }, {
          texture,
          mipLevel: i,
        }, mipSize)

        mipSize.width = mip(mipSize.width)
        mipSize.height = mip(mipSize.height)
      }
    }

    this._device.queue.submit([encoder.finish()])

    if (!renderToSource) {
      tempTexture.destroy()
    }

    return texture
  }

  destroy() {
    this._device = undefined
    this._sampler = undefined
    this._shader = undefined
    this._pipelines = undefined
    this._bindGroupLayout = undefined
    this._pipelineLayout = undefined
  }
}

